from datetime import datetime, timedelta
import functools
import inspect

from requests_sigv4 import Sigv4Request
import boto3
import botocore
from botocore.credentials import RefreshableCredentials
from botocore.exceptions import UnknownServiceError
from dateutil.tz import tzutc


class Session(object):
    """AWS Session for CMDB access."""

    def __init__(
        self,
        role_arn=None,
        aws_access_key_id=None,
        aws_secret_access_key=None,
        region=None,
    ):
        self.region = region
        self.role_arn = role_arn
        self.assumed = None
        self.passed = None
        self.resolved = None
        if all([aws_access_key_id, aws_secret_access_key]):
            self.passed = {'access_key': aws_access_key_id,
                           'secret_key': aws_secret_access_key}
        elif any([aws_access_key_id, aws_secret_access_key]):
            raise ValueError(
                'must provide both aws_access_key_id and aws_secret_key or neither')
        else:
            self._resolve()

        if self.role_arn:
            self.assumed = self._assume_role()

    def _resolve(self):
        """Resolve credentials.

        botocore.credentials.RefreshableCredentials has a _refresh() method, but it only
        works if the credentials have not yet expired yet. So when they do expire we need
        to call this method again to get new credentials.
        """
        session = botocore.session.Session()
        resolver = botocore.credentials.create_credential_resolver(session)
        self.resolved = resolver.load_credentials()
        if not self.resolved:
            raise EnvironmentError('no aws credentials found')

    def _assume_role(self):
        sts_kwargs = {'region_name': self.region}
        if self.passed:
            sts_kwargs['aws_access_key_id'] = self.passed['access_key']
            sts_kwargs['aws_secret_access_key'] = self.passed['secret_key']

        if self.resolved:
            # For Credentials, get_frozen_credentials prevents race conditions
            # when calling self.resolved.access_key|secret_key|token, which each
            # getter in the object attempts to refresh creds. So it is possible
            # to get mismatched credentials.
            frozen_credentials = self.resolved.get_frozen_credentials()
            sts_kwargs['aws_access_key_id'] = frozen_credentials.access_key
            sts_kwargs['aws_secret_access_key'] = frozen_credentials.secret_key
            sts_kwargs['aws_session_token'] = frozen_credentials.token

        sts = boto3.client('sts', **sts_kwargs)
        try:
            return sts.assume_role(RoleArn=self.role_arn,
                                   RoleSessionName='seshypy')
        except botocore.exceptions.ClientError as e:
            msg = 'Assume role failed with the following arguments ' \
                  'region_name: {0}, ' \
                  'aws_access_key_id: {1}, ' \
                  'aws_secret_access_key: {2}, ' \
                  'aws_session_token: {3}'
            e.args += (msg.format(
                sts_kwargs['region_name'],
                sts_kwargs['aws_access_key_id'],
                sts_kwargs['aws_secret_access_key'],
                sts_kwargs['aws_session_token']
            ),)
            raise

    def _verify_resolved(self):
        """Verify or update resolved credentials."""
        if (
            self.resolved and
            self.resolved is RefreshableCredentials and
            self.resolved.refresh_needed()
        ):
            self._resolve()

    def requests(self):
        """Return requester for other sessions."""
        kwargs = {'region': self.region}
        self._verify_resolved()
        if self.assumed:
            reset_time = datetime.now(tz=tzutc()) + timedelta(seconds=30)
            if self.assumed['Credentials']['Expiration'] < reset_time:
                self.assumed = self._assume_role()

            kwargs['access_key'] = self.assumed['Credentials']['AccessKeyId']
            kwargs['secret_key'] = self.assumed['Credentials']['SecretAccessKey']
            kwargs['session_token'] = self.assumed['Credentials']['SessionToken']
            kwargs['session_expires'] = self.assumed['Credentials']['Expiration']
            return Sigv4Request(**kwargs)

        kwargs['access_key'] = (self.passed['access_key']
                                if self.passed else self.resolved.access_key)
        kwargs['secret_key'] = (self.passed['secret_key']
                                if self.passed else self.resolved.secret_key)
        if self.resolved:
            kwargs['session_token'] = self.resolved.token

        return Sigv4Request(**kwargs)

    def __boto3_kwargs(self):
        """Get boto3 client or resource kwargs."""
        kwargs = {'region_name': self.region}
        self._verify_resolved()
        if self.assumed:
            reset_time = datetime.now(tz=tzutc()) + timedelta(seconds=30)
            if self.assumed['Credentials']['Expiration'] < reset_time:
                self.assumed = self._assume_role()
            kwargs['aws_access_key_id'] = self.assumed['Credentials']['AccessKeyId']
            kwargs['aws_secret_access_key'] = self.assumed['Credentials']['SecretAccessKey']
            kwargs['aws_session_token'] = self.assumed['Credentials']['SessionToken']
            return kwargs

        kwargs['aws_access_key_id'] = (
            self.passed['access_key']
            if self.passed else self.resolved.access_key
        )
        kwargs['aws_secret_access_key'] = (
            self.passed['secret_key']
            if self.passed else self.resolved.secret_key
        )
        return kwargs

    def client(self, *args, **kwargs):
        """Get a boto3 client."""
        return self.__persistent_methods(self.__client(*args, **kwargs))

    def resource(self, *args, **kwargs):
        """Get a boto3 resource."""
        return self.__persistent_methods(self.__resource(*args, **kwargs))

    def __client(self, *args, **kwargs):
        """Get a boto3 client."""
        kwargs.update(self.__boto3_kwargs())
        return boto3.client(*args, **kwargs)

    def __resource(self, *args, **kwargs):
        """Get a boto3 resource."""
        kwargs.update(self.__boto3_kwargs())
        return boto3.resource(*args, **kwargs)

    def __persistent_methods(self, obj):
        """Make all methods persistent.

        Args:
            obj (object): boto3.client or boto3.resource
        """
        for name, method in inspect.getmembers(obj, inspect.ismethod):
            setattr(obj, name, self.__persist(method))
        if hasattr(obj.meta, 'client'):
            for name, method in inspect.getmembers(obj.meta.client, inspect.ismethod):
                setattr(obj.meta.client, name, self.__persist(method))
        return obj

    def __persist(self, method):
        """Force all calls on methods to check credentials.

        Args:
            method (method): method to wrap with persistence

        Returns:
            method: method that knows to verify credentials before __call__
        """
        @functools.wraps(method)
        def new_and_call(*args, **kwargs):
            try:
                service = method.__self__.__class__.__name__.lower()
                client = self.__client(service)
            except UnknownServiceError:
                service = method.__self__.__class__.__name__.split('.')[0].lower()
                resource = self.__resource(service)
            try:
                return getattr(client, method.__name__)(*args, **kwargs)
            except UnboundLocalError:
                return getattr(resource, method.__name__)(*args, **kwargs)
        return new_and_call



import json
import os
import sys
import unittest

import pytest
from freezegun import freeze_time

from seshypy.base_session import BaseSession


@pytest.mark.unit
class TestBaseSession(unittest.TestCase):
    @classmethod
    def setup_class(cls):
        os.environ['AWS_ACCESS_KEY_ID'] = 'test'
        os.environ['AWS_DEFAULT_REGION'] = 'us-west-2'
        os.environ['AWS_SECRET_ACCESS_KEY'] = 'test'

    @classmethod
    def teardown_class(cls):
        os.environ.pop('AWS_DEFAULT_REGION', None)
        os.environ.pop('AWS_SECRET_ACCESS_KEY', None)
        os.environ.pop('AWS_ACCESS_KEY_ID', None)

    @freeze_time('2020-02-02')
    def test_base_session_get(self):
        want = {
            'args': {},
            'headers': {
                'Accept': 'application/json',
                'Accept-Encoding': 'identity',
                'Authorization': 'AWS4-HMAC-SHA256 Credential=test/20200202/'
                'us-west-2/execute-api/aws4_request, SignedHeaders=host;'
                'x-amz-date, Signature=db7be03c558c367efbd8cf9d587a6e224'
                'e8380595801586fdeee3223560d7fc4',
                'Content-Type': 'application/json',
                'Host': 'httpbin.org',
                'X-Amz-Date': '20200202T000000Z',
            },
            'url': 'https://httpbin.org/get'
        }
        session = BaseSession(
            host='https://httpbin.org'
        )
        res = session.get(
            path='/get',
        )
        if sys.version_info[0] == 2:
            got = json.loads(res.content)
        else:
            got = json.loads(res.content.decode())

        del got['origin']
        del got['headers']['User-Agent']
        del got['headers']['X-Amzn-Trace-Id']

        assert got == want


if __name__ == '__main__':
    unittest.main()



from cachetools.func import ttl_cache
import logging
import functools

log = logging.getLogger(__name__)


def safe_ttl_cache(*args, **kwargs):
    """Return a ttl_cache wrapped function that can handle unhashable arguments.

    Notes:
        Do this by detecting this condition and returning the results of the
            correct method.

        The outside wrapper, safe_ttl_cache, must capture all arguments to be
            passed to the inner wrapper, wrapped. The mid wrapper,
            func_wrapper, is simply to capture the actual method to be
            fully wrapped.

        safe_ttl_cache shall not be called as a naked decorator. It can be
            called as in examples.

    Examples:
        >>> @safe_ttl_cache()
        >>> def what_you_want_cached(*args, **kwargs)
        >>>     pass

        >>> @safe_ttl_cache(ttl_cache_arguments=here)
        >>> def what_you_want_cached(*args, **kwargs)
        >>>    pass
    """
    def func_wrapper(method):
        ttl_method = ttl_cache(*args, **kwargs)(method)

        @functools.wraps(ttl_method)
        def wrapped(*in_args, **in_kwargs):
            try:
                return ttl_method(*in_args, **in_kwargs)
            except TypeError as err:
                if 'unhashable type' in err.args[0]:
                    log.warning('%s: Returning results without cache. Cannot '
                                'use cache with unhashable type in arguments',
                                method.__name__)
                    return ttl_method.__wrapped__(*in_args, **in_kwargs)
                raise
        return wrapped
    return func_wrapper



import inspect
import logging
try:
    from urllib.parse import urljoin
except ImportError:  # Python 2
    from urlparse import urljoin

import requests
from retrying import retry

from seshypy.session import Session
from seshypy.utils import safe_ttl_cache

log = logging.getLogger(__name__)


class BaseSession(object):
    """BaseSession from which each session derives.

    This object handles authentication (and keeping it alive), and requests.
    It uses the Session object for session handling.

    Args:
        host (str): api endpoint
        role_arn (optional[str]): arn for role to sts into
        aws_access_key_id (optional[str]): access key
        aws_secret_access_key (optional[str]): secret key
        region (optional[str]): api region (required if using access and secret keys)
        session (optional(session)): session to be shared across subclasses
        use_cache (optional[bool]): Use cache? Defaults to False
        cache_ttl (optional[int]): Duration cache lives. Defaults to 1200.
        cache_methods (optional[int]): List of methods to cache.
            Defaults provided by the inheriting classes.

    Returns:
        session instance

    Cache:
        The caching is provided by cachetools_. The defaults provided by each class
        or a list of method names passed in are wrapped by the ttl_cache method. By
        default this sets a ttl of 20 minutes, but this can be configured by passing
        in cache_ttl with the number of seconds you want the cache to live.

        You can disable cache entirely by passing in ``use_cache=False``.

        >>> cmdb = CMDBSession(use_cache=False, **cfg)

        Cache can be invalidated by calling the method ``cache_clear()`` on a given method.

        >>> cmdb.accounts.get_account.cache_clear()

    .. _cachetools:
        http://pythonhosted.org/cachetools/
    """

    def __init__(self, host, role_arn=None, aws_access_key_id=None,
                 aws_secret_access_key=None, region=None, session=None,
                 use_cache=True, cache_ttl=1200, cache_methods=None):
        self.host = host
        self.headers = {
            'Accept': 'application/json',
            'Content-type': 'application/json',
        }
        if not session:
            session = Session(
                role_arn=role_arn,
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                region=region)
        self.session = session
        self.__use_cache = bool(use_cache)
        self.__cache_ttl = cache_ttl
        self.__cache_methods = cache_methods if cache_methods is not None else []
        if self.use_cache:
            for method_name in set(self.cache_methods):
                self.__enable_method_cache(method_name)

    def __enable_method_cache(self, method_name):
        """Enable cache for a method.

        Args:
            method_name (str): name of method for which to enable cache

        Returns:
            bool: True is success, False if failed
        """
        try:
            method = getattr(self, method_name)
            try:
                return not inspect.isfunction(method.cache_info)
            except AttributeError:
                setattr(
                    self, method_name,
                    safe_ttl_cache(ttl=self.cache_ttl)(method))
                return True
        except AttributeError:
            # method doesn't exist on initializing class
            return False

    @property
    def use_cache(self):
        """Boolean to use cache or not."""
        return self.__use_cache

    @property
    def cache_ttl(self):
        """Duration cache lives."""
        return self.__cache_ttl

    @property
    def cache_methods(self):
        """List of methods to cache."""
        return self.__cache_methods

    def make_api_gateway_call(self, method, path, headers=None, qstr_vals=None,
                              params=None, host=None, payload=None):
        """Make call to API Gateway.

        Args:
            method (str): HTTP method
            path (str): request path
            headers (optional[dict]): headers if provided else self.headers
            qstr_vals (optional[dict]): query string parameters to add
                deprecated: use params
            params (optional[dict]): query string parameters
            host (optional[str]): URL to host for request else self.host
            payload (optional[dict]): payload as dictionary

        Returns:
            response object
        """
        # This is to support older implementations using qstr_vals instead of params.
        params = params if params is not None else qstr_vals
        host = host if host else self.host
        headers = headers if headers else self.headers
        req = self.session.requests()
        url = urljoin(host, path)
        try:
            if method == 'POST':
                res = req.post(url, headers=headers,
                               json=payload, params=params)
            elif method == 'GET':
                res = req.get(url, headers=headers, params=params)
            elif method == 'PUT':
                res = req.put(url, headers=headers,
                              params=params, json=payload)
            elif method == 'DELETE':
                res = req.delete(url, headers=headers)
            else:
                raise ValueError('HTTP method {} not supported'.format(method))
            if res.status_code != 200 or not res.json():
                if res.status_code == 200 and not res.json() and method == 'GET':
                    # results are falsy; probably an empty list
                    extra_info = ('{} actually received 200, but the result was empty.'
                                  .format(__package__))
                    res.reason = 'Not Found; {}'.format(extra_info)
                    res.status_code = 404
                log.info(res)
                res.raise_for_status()
            log.info('request to %s', url)
            return res
        except requests.exceptions.RequestException as err:
            log.error(err)
            raise

    def get(self, path, host=None, headers=None, params=None):
        """GET from API Gateway.

        Args:
            path (str): request path
            headers (optional[dict]): headers if provided else self.headers
            params (optional[dict]): query string parameters
            host (optional[str]): URL to host for request else self.host

        Returns:
            response object
        """
        return self.make_api_gateway_call(
            'GET',
            path=path, host=host, headers=headers, params=params
        )

    def post(self, path, host=None, headers=None, params=None, payload=None):
        """POST to API Gateway.

        Args:
            path (str): request path
            host (optional[str]): URL to host for request else self.host
            headers (optional[dict]): headers if provided else self.headers
            params (optional[dict]): query string parameters
            payload (optional[dict]): payload as dictionary

        Returns:
            response object
        """
        return self.make_api_gateway_call(
            'POST',
            path=path, host=host, headers=headers, params=params, payload=payload
        )

    def put(self, path, host=None, headers=None, params=None, payload=None):
        """PUT to API Gateway.

        Args:
            path (str): request path
            host (optional[str]): URL to host for request else self.host
            headers (optional[dict]): headers if provided else self.headers
            params (optional[dict]): query string parameters
            payload (optional[dict]): payload as dictionary

        Returns:
            response object
        """
        return self.make_api_gateway_call(
            'PUT',
            path=path, host=host, headers=headers, params=params, payload=payload
        )

    def delete(self, path, host=None, headers=None):
        """DELETE from API Gateway.

        Args:
            path (str): request path
            host (optional[str]): URL to host for request else self.host
            headers (optional[dict]): headers if provided else self.headers

        Returns:
            response object
        """
        return self.make_api_gateway_call(
            'DELETE',
            path=path, host=host, headers=headers
        )

    def api_retrying(self, *args, **kwargs):
        """Wrap api calls with retrying.

        Args:
            all options provided by retrying library
                url: https://pypi.python.org/pypi/retrying

        Example::

            retrying_kwargs = {
                'retry_on_exception': lambda x: isinstance(x, ConnectionError),
                'stop_max_attempt_number': 3,
                'wait_exponential_multiplier': 1000,
                'wait_exponential_max': 10000
            }
            session.api_retrying(**retrying_kwargs)
        """
        self.make_api_gateway_call = retry(
            *args, **kwargs)(self.make_api_gateway_call)

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        pass



#!/usr/bin/env python
from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

install_requires = [
    'requests-sigv4>=0.1.4',
    'cachetools>=1.1.6',
    'figgypy>=0.2.0',
    'future',
    'requests>=2.21.0',
    'retrying>=1.3.3',
    'setuptools>=36.5.0'
]

setup(
    name='seshypy',
    version='0.6.1',
    description='seshypy makes API Gateway requests and API Gateway clients easy.',
    long_description=readme,
    packages=find_packages(),
    platforms=['all'],
    install_requires=install_requires,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)
